import mock
from click.testing import CliRunner

from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import example_responses, MockResponse

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestListProjects(object):
    URL = "https://api.paperspace.io/projects/"
    EXPECTED_PARAMS = {
        "filter": """{"offset":0,"where":{"dtDeleted":null},"order":"dtCreated desc"}"""
    }
    EXPECTED_PARAMS_WITH_FILTERING_BY_TAGS = {
        "filter": """{"offset":0,"where":{"dtDeleted":null},"order":"dtCreated desc"}""",
        "tagFilter": ("some_tag", "some_tag_2"),
    }
    BASIC_COMMAND = ["projects", "list"]
    BASIC_COMMAND_WITH_FILTERING_BY_TAGS = ["projects", "list", "--tag", "some_tag", "--tag", "some_tag_2"]
    # TODO: change to `REQUEST_JSON = None` or whatever works when PS_API is fixed
    EXPECTED_RESPONSE_JSON = example_responses.LIST_PROJECTS_RESPONSE
    EXPECTED_STDOUT = """+-----------+-------------------+------------+----------------------------+
| ID        | Name              | Repository | Created                    |
+-----------+-------------------+------------+----------------------------+
| prq70zy79 | test_project      | None       | 2019-03-18 13:24:46.666000 |
| prmr22ve0 | keton             | None       | 2019-03-25 14:50:43.202000 |
| przhbct98 | paperspace-python | None       | 2019-04-04 15:12:34.229000 |
+-----------+-------------------+------------+----------------------------+
"""

    BASIC_COMMAND_WITH_API_KEY = ["projects", "list", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE = ["projects", "list", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    RESPONSE_JSON_WHEN_NO_PROJECTS_WERE_FOUND = {"data": [], "meta": {"totalItems": 0}}
    EXPECTED_STDOUT_WHEN_NO_PROJECTS_WERE_FOUND = "No data found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_projects_list_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_projects_list_was_used_with_filtering_by_tags(
            self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_FILTERING_BY_TAGS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITH_FILTERING_BY_TAGS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_projects_list_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, projects_list_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [projects_list_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_projects_list_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_no_project_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_NO_PROJECTS_WERE_FOUND,
                                                status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WHEN_NO_PROJECTS_WERE_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0


class TestCreateProject(object):
    URL = "https://api.paperspace.io/projects/"
    TAGS_URL = "https://api.paperspace.io/entityTags/updateTags"
    COMMAND = [
        "projects", "create",
        "--name", "some_name",
    ]
    COMMAND_WITH_TAGS = [
        "projects", "create",
        "--name", "some_name",
        "--tag", "test0",
        "--tag", "test1",
        "--tags", "test2,test3",
    ]
    EXPECTED_REQUEST_JSON = {"name": "some_name"}
    TAGS_JSON = {
        "entity": "project",
        "entityId": "pru5a4dnu",
        "tags": ["test0", "test1", "test2", "test3"]
    }
    EXPECTED_RESPONSE_JSON = {
        "name": "some_name",
        "handle": "pru5a4dnu",
        "dtCreated": "2019-06-10T12:20:17.173Z",
        "dtDeleted": None,
        "lastJobSeqNum": 0,
        "repoNodeId": None,
        "repoName": None,
        "repoUrl": None,
    }

    COMMAND_WHEN_ALL_PARAMETERS_WERE_USED = [
        "projects", "create",
        "--name", "some_name",
        "--repositoryName", "some_repository_name",
        "--repositoryUrl", "some_repository_url",
    ]
    COMMAND_WITH_OPTIONS_FILE = ["projects", "create", "--optionsFile", ]  # path added in test

    EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED = {
        "repoUrl": "some_repository_url",
        "name": "some_name",
        "repoName": "some_repository_name",
    }
    EXPECTED_RESPONSE_JSON_WHEN_ALL_PARAMETERS_WERE_USED = {
        "name": "mnist",
        "handle": "pru5a4dnu",
        "dtCreated": "2019-06-10T12:25:59.464Z",
        "dtDeleted": None,
        "lastJobSeqNum": 0,
        "repoNodeId": None,
        "repoName": "mnist-sample",
        "repoUrl": "https://github.com/Paperspace/mnist-sample",
    }
    UPDATE_TAGS_RESPONSE_JSON_200 = example_responses.UPDATE_TAGS_RESPONSE

    EXPECTED_STDOUT = "Project created with ID: pru5a4dnu\n" \
                      "https://console.paperspace.com/projects/pru5a4dnu\n"

    EXPECTED_RESPONSE_JSON_WITH_ERROR = {
        "error": {
            "name": "ValidationError",
            "status": 422,
            "message": "The `project` instance is not valid. Details: `name` can't be blank (value: undefined).",
            "statusCode": 422,
            "details": {
                "context": "project",
                "codes": {"name": ["presence"]},
                "messages": {"name": ["can't be blank"]},
            },
        },
    }
    EXPECTED_STDOUT_WHEN_ERROR_RECEIVED = "Failed to create resource: The `project` instance is not valid. " \
                                          "Details: `name` can't be blank (value: undefined).\n"

    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "projects", "create",
        "--name", "some_name",
        "--apiKey", "some_key",
    ]

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED = "Failed to create resource: Invalid API token\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_and_print_proper_message_when_create_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_ALL_PARAMETERS_WERE_USED, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_and_print_proper_message_when_create_command_was_used_with_all_options(
            self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_ALL_PARAMETERS_WERE_USED, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WHEN_ALL_PARAMETERS_WERE_USED)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON, 201)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, projects_create_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON, 201)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [projects_create_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON_WHEN_ALL_PARAMETERS_WERE_USED,
                                             params=None,
                                             files=None,
                                             data=None)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_error_message_received(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WITH_ERROR, 400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_ERROR_RECEIVED

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 403)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == self.EXPECTED_STDOUT_WHEN_WRONG_API_KEY_WAS_USED
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_request_and_print_proper_message_when_error_code_returned_without_json_data(self,
                                                                                                     post_patched):
        post_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        assert result.output == "Failed to create resource\n"
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_tag_project(self, post_patched, get_patched, put_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_ALL_PARAMETERS_WERE_USED, 201)
        get_patched.return_value = MockResponse({}, 200, "fake content")
        put_patched.return_value = MockResponse(self.UPDATE_TAGS_RESPONSE_JSON_200, 200, "fake content")

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_TAGS)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        put_patched.assert_called_once_with(
            self.TAGS_URL,
            headers=EXPECTED_HEADERS,
            json=self.TAGS_JSON,
            params=None,
            data=None,
        )

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        assert result.exit_code == 0


class TestDeleteProjects(object):
    URL = "https://api.paperspace.io/projects/some_project_id/deleteProject"
    BASIC_COMMAND = ["projects", "delete", "--id", "some_project_id"]
    EXPECTED_STDOUT = "Project deleted\n"

    BASIC_COMMAND_WITH_API_KEY = ["projects", "delete", "--id", "some_project_id", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE = ["projects", "delete",
                                 "--id", "some_project_id",
                                 "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to delete resource: Invalid API token\n"

    RESPONSE_JSON_WHEN_PROJECT_IS_ALREADY_DELETED = {
        "error": {
            "name": "Error",
            "status": 400,
            "message": "Project \"some_project_id\" is already deleted.",
        },
    }
    EXPECTED_STDOUT_WHEN_PROJECT_IS_ALREADY_DELETED = \
        """Failed to delete resource: Project "some_project_id" is already deleted.\n"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_and_print_valid_message_when_delete_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=204)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        data=None,
                                        files=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_delete_was_used_with_api_key_option(self, post_patched):
        post_patched.return_value = MockResponse(status_code=204)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        data=None,
                                        files=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, projects_delete_config_path):
        post_patched.return_value = MockResponse(status_code=204)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [projects_delete_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        data=None,
                                        files=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_delete_was_used_with_wrong_api_key(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        data=None,
                                        files=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_the_project_is_already_deleted(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_PROJECT_IS_ALREADY_DELETED,
                                                 status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        data=None,
                                        files=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_PROJECT_IS_ALREADY_DELETED
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_project_was_not_found(self, post_patched):
        # TODO: add test later when the API is fixed - it responds with invalid message
        pass

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=None,
                                       data=None,
                                       files=None)
        assert result.output == "Failed to delete resource\n"
        assert result.exit_code == 0


class TestProjectsDetails(object):
    URL = "https://api.paperspace.io/projects/"
    EXPECTED_PARAMS = {
        "filter": """{"where":{"handle":"some_id"}}"""
    }
    BASIC_COMMAND = ["projects", "details", "--id", "some_id"]
    COMMAND_WITH_OPTIONS_FILE = ["projects", "details", "--id", "some_id", "--optionsFile", ]  # path added in test
    BASIC_COMMAND_WITH_API_KEY = ["projects", "details", "--id", "some_id", "--apiKey", "some_key"]

    EXPECTED_RESPONSE_JSON = example_responses.DETAILS_OF_PROJECT
    EXPECTED_STDOUT = """+-----------------+-----------+
| Name            | some_name |
+-----------------+-----------+
| ID              | some_id   |
| Repository name | None      |
| Repository url  | None      |
| Tags            |           |
+-----------------+-----------+
"""

    EXPECTED_RESPONSE_JSON_WITH_TAGS = example_responses.DETAILS_OF_PROJECT_WITH_TAGS
    EXPECTED_STDOUT_WITH_TAGS = """+-----------------+------------+
| Name            | some_name  |
+-----------------+------------+
| ID              | some_id    |
| Repository name | None       |
| Repository url  | None       |
| Tags            | tag1, tag2 |
+-----------------+------------+
"""

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    RESPONSE_JSON_WHEN_PROJECT_WAS_NOT_FOUND = {
        "data": [],
        "meta": {
            "where": {"handle": "some_id"},
            "totalItems": 0,
        },
        "tagFilter": [],
    }
    EXPECTED_STDOUT_WHEN_PROJECT_WAS_NOT_FOUND = "Project not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_projects_details_was_run(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_projects_details_was_run_with_tags(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON_WITH_TAGS)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WITH_TAGS, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_projects_details_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON_WITH_TAGS)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_TAGS
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, projects_details_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON_WITH_TAGS)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [projects_details_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT_WITH_TAGS, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_projects_details_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_project_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_PROJECT_WAS_NOT_FOUND)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WHEN_PROJECT_WAS_NOT_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0



import abc
import json
import os

import six
import yaml

from halo import halo
from gradient import api_sdk
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.exceptions import ApplicationError
from gradient.commands.common import BaseCommand, ListCommandMixin, DetailsCommandMixin, LogsCommandMixin


@six.add_metaclass(abc.ABCMeta)
class DetailJSONCommandMixin(DetailsCommandMixin):
    def _get_table_data(self, instance):
        json_formatted_str = json.dumps(instance, indent=4)
        return json_formatted_str

    def _log_object(self, instance):
        if instance is None:
            self.logger.warning("Not found")
            return
        
        json_formatted_str = json.dumps(instance, indent=4)
        print(json_formatted_str)


@six.add_metaclass(abc.ABCMeta)
class BaseWorkflowCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.WorkflowsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class CreateWorkflowCommand(BaseWorkflowCommand, DetailsCommandMixin):
    def _get_table_data(self, instance):
        data = [('Name', 'ID')]
        data.append((
            instance['name'],
            instance['id'],
        ))

        return data

    def execute(self, name, project_id):
        workflow = self.client.create(name, project_id)
        self._log_object(workflow)

        return workflow

class CreateWorkflowRunCommand(BaseWorkflowCommand):
    def execute(self, spec_path=None, input_path=None, workflow_id=None, cluster_id=None):
        spec = None
        inputs = None
        if spec_path:
            if not os.path.exists(spec_path):
                raise ApplicationError('Source path not found: {}'.format(spec_path))
            else:
                yaml_spec = open(spec_path, 'r')
                spec = yaml.safe_load(yaml_spec)

        if input_path:
            if not os.path.exists(input_path):
                raise ApplicationError('Source path not found: {}'.format(input_path))
            else:
                yaml_inputs = open(input_path, 'r')
                inputs = yaml.safe_load(yaml_inputs)

        workflow = self.client.run_workflow(spec=spec, inputs=inputs, workflow_id=workflow_id, cluster_id=cluster_id)
        return workflow


class ListWorkflowsCommand(ListCommandMixin, BaseWorkflowCommand):
    def _get_instances(self, kwargs):
        instances = self.client.list(**kwargs)
        return instances

    def _get_table_data(self, objects):
        data = [('Name', 'ID')]
        for workflow in objects:
            data.append((
                workflow.name,
                workflow.id,
            ))
        return data

class ListWorkflowRunsCommand(ListCommandMixin, BaseWorkflowCommand):
    def _get_instances(self, kwargs):
        instances = self.client.list_runs(**kwargs)
        return instances
        
    def _get_table_data(self, objects):
        data = [('Run', 'Cluster ID', 'Status')]
        for workflow_run in objects:
            data.append((
                workflow_run['id'],
                workflow_run['cluster']['id'],
                workflow_run['status']['phase'],
            ))
        return data


class GetWorkflowCommand(DetailJSONCommandMixin, BaseWorkflowCommand):
    def execute(self, workflow_id):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instance = self.get_instance(workflow_id=workflow_id)

        self._log_object(instance)

    def get_instance(self, workflow_id):
        instance = self.client.get(workflow_id=workflow_id)
        return instance

class GetWorkflowRunCommand(DetailJSONCommandMixin, BaseWorkflowCommand):
    def execute(self, workflow_id, run):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instance = self.get_instance(workflow_id=workflow_id, run=run)

        self._log_object(instance)

    def get_instance(self, workflow_id, run):
        instance = self.client.get_run(workflow_id=workflow_id, run=run)
        return instance

class WorkflowLogsCommand(LogsCommandMixin, BaseWorkflowCommand):
    ENTITY = "Workflows"




import gradient.api_sdk.config
from .. import serializers, s3_uploader
from ..repositories.common import ListResources, DeleteResource, CreateResource, GetResource
from ..sdk_exceptions import ResourceFetchingError


class ParseModelDictMixin(object):
    def _parse_object(self, model_dict, **kwargs):
        """
        :param dict model_dict:
        :rtype Model
        """
        model = serializers.Model().get_instance(model_dict)
        return model


class GetBaseModelsApiUrlMixin(object):
    def _get_api_url(self, **_):
        return gradient.api_sdk.config.config.CONFIG_HOST


class ListModels(GetBaseModelsApiUrlMixin, ParseModelDictMixin, ListResources):
    def get_request_url(self, **kwargs):
        return "/mlModels/getModelList/"

    def _parse_objects(self, data, **kwargs):
        experiments = []
        for model_dict in data["modelList"]:
            experiment = self._parse_object(model_dict)
            experiments.append(experiment)

        return experiments

    def _get_request_params(self, kwargs):
        return {"limit": -1}

    def _get_request_json(self, kwargs):
        filters = {}
        if kwargs.get("experiment_id"):
            filters["experimentId"] = kwargs.get("experiment_id")
        if kwargs.get("project_id"):
            filters["projectId"] = kwargs.get("project_id")

        if filters:
            json_ = {"filter": {"where": {"and": [filters]}}}
        else:
            json_ = {}

        tags = kwargs.get("tags")
        if tags:
            json_["tagFilter"] = tags

        return json_ or None


class DeleteModel(GetBaseModelsApiUrlMixin, DeleteResource):
    def get_request_url(self, **kwargs):
        return "/mlModels/deleteModel/"

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response

    def _get_request_json(self, kwargs):
        return kwargs


class GetPresignedUrlForModelFile(GetResource):
    def get_request_url(self, **kwargs):
        return "/mlModels/createModelV2"

    def _get_api_url(self, **kwargs):
        return gradient.api_sdk.config.config.CONFIG_HOST


class UploadModel(GetBaseModelsApiUrlMixin, CreateResource):
    SERIALIZER_CLS = serializers.Model
    HANDLE_FIELD = "id"

    def get_request_url(self, **kwargs):
        return "/mlModels/createModelV2"

    def _get_request_params(self, kwargs):
        return kwargs

    def _get_request_json(self, instance_dict):
        return None

    def create(self, instance, data=None, path=None, cluster_id=None):
        model_id = super(UploadModel, self).create(instance, data=data, path=path)
        try:
            self._upload_model(path, model_id, cluster_id=cluster_id)
        except BaseException:
            self._delete_model(model_id)
            raise

        return model_id

    def _upload_model(self, file_path, model_id, cluster_id=None):
        model_uploader = s3_uploader.S3ModelUploader(
            self.api_key, logger=self.logger, ps_client_name=self.ps_client_name
        )
        model_uploader.upload(file_path, model_id, cluster_id=cluster_id)

    def _delete_model(self, model_id):
        repository = DeleteModel(self.api_key, logger=self.logger, ps_client_name=self.ps_client_name)
        repository.delete(model_id)


class CreateModel(GetBaseModelsApiUrlMixin, CreateResource):
    SERIALIZER_CLS = serializers.Model
    HANDLE_FIELD = "id"

    def get_request_url(self, **kwargs):
        return "/mlModels/createModelV2"

    def _get_request_params(self, kwargs):
        return kwargs

    def _get_request_json(self, instance_dict):
        return None

    def create(self, instance, data=None, path=None):
        model_id = super(CreateModel, self).create(instance, data=data, path=path)
        return model_id


class GetModel(GetBaseModelsApiUrlMixin, GetResource):
    SERIALIZER_CLS = serializers.Model

    def get_request_url(self, **kwargs):
        return "/mlModels/getModelList/"

    def _get_request_json(self, kwargs):
        model_id = kwargs["model_id"]
        json_ = {"filter": {"where": {"and": [{"id": model_id}]}}}
        return json_

    def _parse_object(self, instance_dict, **kwargs):
        try:
            model_dict = instance_dict["modelList"][0]
        except (IndexError, TypeError):
            raise ResourceFetchingError("Model not found")

        instance = self.SERIALIZER_CLS().get_instance(model_dict)
        return instance


class ListModelFiles(GetBaseModelsApiUrlMixin, ListResources):
    SERIALIZER_CLS = serializers.ModelFileSchema

    def get_request_url(self, **kwargs):
        return "/mlModels/listFiles/"

    def _get_request_json(self, kwargs):
        json_ = {"id": kwargs["model_id"]}
        if kwargs.get("links"):
            json_["links"] = True
        if kwargs.get("size"):
            json_["size"] = True

        return json_



import abc

import halo
import six

from gradient import api_sdk, exceptions
from gradient.api_sdk import sdk_exceptions
from gradient.api_sdk.config import config
from gradient.api_sdk.utils import concatenate_urls
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from .common import BaseCommand, ListCommandMixin, DetailsCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseProjectCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.ProjectsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class CreateProjectCommand(BaseProjectCommand):
    SPINNER_MESSAGE = "Creating new project"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "Project created with ID: {}"

    def execute(self, project_dict):
        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            project_id = self.client.create(**project_dict)

        self.logger.log(self.CREATE_SUCCESS_MESSAGE_TEMPLATE.format(project_id))
        self.logger.log(self.get_instance_url(project_id))

    def get_instance_url(self, project_id):
        url = concatenate_urls(config.WEB_URL, "{}/projects/{}".format(self.get_namespace(), project_id))
        return url


class ListProjectsCommand(ListCommandMixin, BaseProjectCommand):
    def _get_instances(self, kwargs):
        try:
            instances = self.client.list(**kwargs)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances

    def _get_table_data(self, objects):
        data = [("ID", "Name", "Repository", "Created")]
        for obj in objects:
            created_str = obj.created.strftime("%Y-%m-%d %H:%M:%S.%f")
            data.append((obj.id, obj.name, obj.repository_url, created_str))

        return data


class DeleteProjectCommand(BaseProjectCommand):
    def execute(self, project_id):
        self.client.delete(project_id)
        self.logger.log("Project deleted")


class ProjectAddTagsCommand(BaseProjectCommand):
    def execute(self, project_id, *args, **kwargs):
        self.client.add_tags(project_id, **kwargs)
        self.logger.log("Tags added to project")


class ProjectRemoveTagsCommand(BaseProjectCommand):
    def execute(self, project_id, *args, **kwargs):
        self.client.remove_tags(project_id, **kwargs)
        self.logger.log("Tags removed from project")


class ShowProjectDetailsCommand(DetailsCommandMixin, BaseProjectCommand):
    def _get_table_data(self, instance):
        """
        :param api_sdk.Project instance:
        """
        tags_string = ", ".join(instance.tags)

        data = (
            ("Name", instance.name),
            ("ID", instance.id),
            ("Repository name", instance.repository_name),
            ("Repository url", instance.repository_url),
            ("Tags", tags_string),
        )
        return data



from .. import repositories

from ..clients.base_client import BaseClient


class MachineTypesClient(BaseClient):
    def list(self, cluster_id=None):
        repository = self.build_repository(repositories.ListMachineTypes)
        machine_types = repository.list(cluster_id=cluster_id)
        return machine_types



from .base_client import BaseClient
from .. import models, repositories
from ...exceptions import ReceivingDataFailedError


class WorkflowsClient(BaseClient):

    def create(self, name, project_id):
        """Create workflow with spec 

        :param str name: workflow name
        :param str project_id: project id

        :returns:  workflow create response
        :rtype: list[models.Workflow]
        """

        repository = self.build_repository(repositories.CreateWorkflow)
        workflow = repository.create(name=name, project_id=project_id)
        return workflow

    def run_workflow(self, spec, inputs, workflow_id, cluster_id):
        """Create workflow with spec 

        :param obj spec: workflow spec
        :param obj inputs: workflow inputs
        :param str workflow_id: workflow id
        :param str cluster_id: cluster id

        :returns:  workflow create response
        :rtype: list[models.Workflow]
        """

        repository = self.build_repository(repositories.CreateWorkflowRun)
        workflow = repository.create(spec=spec, inputs=inputs, id=workflow_id, cluster_id=cluster_id)
        return workflow

    def list(self, project_id):
        """List workflows by project

        :param str project_id: project ID

        :returns: list of workflows
        :rtype: list[models.Workflow]
        """

        repository = self.build_repository(repositories.ListWorkflows)
        workflows = repository.list(project_id=project_id)
        return workflows

    def get(self, workflow_id):
        """Get a Workflow

        :param str workflow_id: Workflow ID [required]

        :returns: workflow
        :rtype: models.Workflow
        """
        repository = self.build_repository(repositories.GetWorkflow)
        return repository.get(id=workflow_id)
    

    def list_runs(self, workflow_id):
        """List workflows runs by workflow id

        :param str workflow_id: workflow ID

        :returns: list of workflow runs
        """

        repository = self.build_repository(repositories.ListWorkflowRuns)
        workflows_runs = repository.get(id=workflow_id)
        return workflows_runs

    def get_run(self, workflow_id, run):
        """List workflows runs by workflow id

        :param str workflow_id: workflow ID
        :param str run: run count

        :returns: list of workflow runs
        """

        repository = self.build_repository(repositories.GetWorkflowRun)
        workflows_runs = repository.get(id=workflow_id, run=run)
        return workflows_runs

    def yield_logs(self, job_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_logs_generator = job_client.yield_logs(
                job_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str job_id:
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        repository = self.build_repository(repositories.ListWorkflowLogs)
        logs = repository.yield_logs(id=job_id, line=line, limit=limit)
        return logs
    
    def logs(self, job_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_logs_generator = job_client.yield_logs(
                job_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str job_id:
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        repository = self.build_repository(repositories.ListWorkflowLogs)
        logs = repository.list(id=job_id, line=line, limit=limit)
        return logs


import mock
from click.testing import CliRunner

from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import MockResponse, example_responses

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestMachineAvailability(object):
    URL = "https://api.paperspace.io/machines/getAvailability/"
    COMMAND = [
        "machines", "availability",
        "--region", "ny2",
        "--machineType", "P4000",
    ]
    PARAMS = {"region": "East Coast (NY2)", "machineType": "P4000"}
    RESPONSE_JSON = {"available": True}

    COMMAND_WITH_API_KEY = [
        "machines", "availability",
        "--region", "ny2",
        "--machineType", "P4000",
        "--apiKey", "some_key",
    ]
    COMMAND_WITH_OPTIONS_FILE = ["machines", "availability", "--optionsFile", ]  # path added in test

    EXPECTED_STDOUT = "Machine available: True\n"

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_valid_message_when_availability_command_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_availability_config_path):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_availability_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_valid_error_message_when_availability_command_was_used_with_invalid_api_token(self,
                                                                                                        get_patched):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, 400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_valid_error_message_when_no_content_was_received_in_response(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.PARAMS)
        assert result.output == "Unknown error while checking machine availability\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_valid_error_message_when_no_content_was_received_in_response(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0


class TestCreateMachine(object):
    URL = "https://api.paperspace.io/machines/createSingleMachinePublic/"
    TAGS_URL = "https://api.paperspace.io/entityTags/updateTags"
    BASIC_COMMAND = [
        "machines", "create",
        "--region", "CA1",
        "--machineType", "P5000",
        "--size", 2,
        "--billingType", "hourly",
        "--machineName", "some_name",
        "--templateId", "some_template",
    ]
    BASIC_COMMAND_WITH_TAGS = [
        "machines", "create",
        "--region", "CA1",
        "--machineType", "P5000",
        "--size", 2,
        "--billingType", "hourly",
        "--machineName", "some_name",
        "--templateId", "some_template",
        "--tag", "test0",
        "--tag", "test1",
        "--tags", "test2,test3",
    ]
    REQUEST_JSON = {
        "billingType": "hourly",
        "machineType": "P5000",
        "machineName": "some_name",
        "region": "West Coast (CA1)",
        "templateId": "some_template",
        "size": 2,
    }
    TAGS_JSON = {
        "entity": "machine",
        "entityId": "psclbvqpc",
        "tags": ["test0", "test1", "test2", "test3"]
    }

    ALL_COMMANDS = [
        "machines", "create",
        "--apiKey", "some_key",
        "--assignPublicIp",
        "--billingType", "hourly",
        "--email", "some@ema.il",
        "--firstName", "some_f_name",
        "--lastName", "some_l_name",
        "--machineName", "some_name",
        "--machineType", "P5000",
        "--networkId", "some_network_id",
        "--notificationEmail", "some@em.a.il",
        "--password", "some_password",
        "--region", "CA1",
        "--scriptId", "some_script_id",
        "--size", "2",
        "--teamId", "some_team_id",
        "--templateId", "some_template",
    ]
    ALL_OPTIONS_REQUEST_JSON = {
        "assignPublicIp": True,
        "billingType": "hourly",
        "email": "some@ema.il",
        "firstName": "some_f_name",
        "lastName": "some_l_name",
        "machineName": "some_name",
        "machineType": "P5000",
        "networkId": "some_network_id",
        "notificationEmail": "some@em.a.il",
        "password": "some_password",
        "region": "West Coast (CA1)",
        "scriptId": "some_script_id",
        "size": 2,
        "teamId": "some_team_id",
        "templateId": "some_template",
    }

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "create",
        "--region", "CA1",
        "--machineType", "P5000",
        "--size", 2,
        "--billingType", "hourly",
        "--machineName", "some_name",
        "--templateId", "some_template",
        "--apiKey", "some_key",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["machines", "create", "--optionsFile", ]  # path added in test

    EXPECTED_STDOUT = "New machine created with id: psclbvqpc\n" \
                      "https://console.paperspace.com/machines/psclbvqpc\n"

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to create resource: Invalid API token\n"
    UPDATE_TAGS_RESPONSE_JSON_200 = example_responses.UPDATE_TAGS_RESPONSE

    RESPONSE_JSON_WITH_WRONG_TEMPLATE_ID = {
        "error": {
            "name": "Error",
            "status": 400,
            "message": "templateId not found"
        }
    }

    BASIC_COMMAND_WITH_MUTUALLY_EXCLUSIVE_OPTIONS_USED = [
        "machines", "create",
        "--region", "CA1",
        "--machineType", "P5000",
        "--size", 2,
        "--billingType", "hourly",
        "--machineName", "some_name",
        "--templateId", "some_template",
        "--userId", "some_user_id",
        "--email", "some@email.com",
    ]

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machine_create_was_used_with_requested_options(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machine_create_was_used_with_all_options(self, post_patched):
        post_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.ALL_COMMANDS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=self.ALL_OPTIONS_REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=self.REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, machines_create_config_path):
        post_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_create_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=self.ALL_OPTIONS_REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, 400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=self.REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_wrong_template_id_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_TEMPLATE_ID, 400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=self.REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Failed to create resource: templateId not found\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_no_content_was_received_in_response(self, post_patched):
        post_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=self.REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Failed to create resource\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_mutually_exclusive_options_were_used(self, get_patched):
        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_MUTUALLY_EXCLUSIVE_OPTIONS_USED)

        get_patched.assert_not_called()
        assert "Error: --userId is mutually exclusive with --email, --password, --firstName and --lastName\n" \
               in result.output
        assert result.exit_code == 2

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_tag_machine(self, post_patched, get_patched, put_patched):
        post_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)
        get_patched.return_value = MockResponse({}, 200)
        put_patched.return_value = MockResponse(self.UPDATE_TAGS_RESPONSE_JSON_200, 200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_TAGS)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.REQUEST_JSON,
                                             params=None,
                                             files=None,
                                             data=None)

        put_patched.assert_called_once_with(
            self.TAGS_URL,
            headers=EXPECTED_HEADERS,
            json=self.TAGS_JSON,
            params=None,
            data=None,
        )

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        assert result.exit_code == 0


class TestDestroyMachine(object):
    URL = "https://api.paperspace.io/machines/some_id/destroyMachine/"
    BASIC_COMMAND = [
        "machines", "destroy",
        "--id", "some_id",
    ]

    ALL_COMMANDS = [
        "machines", "destroy",
        "--id", "some_id",
        "--releasePublicIp",
    ]
    ALL_COMMANDS_REQUEST_JSON = {"releasePublicIp": True}

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "destroy",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["machines", "destroy", "--optionsFile", ]  # path added in test

    EXPECTED_STDOUT = "Machine successfully destroyed\n"

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to delete resource: Invalid API token\n"

    RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Not found. Please contact support@paperspace.com for help."
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Failed to delete resource: Not found. " \
                                                 "Please contact support@paperspace.com for help.\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machines_destroy_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machines_destroy_was_used_with_all_options(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.ALL_COMMANDS)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=self.ALL_COMMANDS_REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, machines_destroy_config_path):
        post_patched.return_value = MockResponse()
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_destroy_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=self.ALL_COMMANDS_REQUEST_JSON,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machines_destroy_was_used_with_api_key_option(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machines_destroy_was_used_with_wrong_api_key(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_machine_with_given_id_was_not_found(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND,
                                                 status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, post_patched):
        post_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Failed to delete resource\n"
        assert result.exit_code == 0


class TestListMachines(object):
    URL = "https://api.paperspace.io/machines/getMachines/"
    BASIC_COMMAND = ["machines", "list"]
    REQUEST_JSON = {}
    EXPECTED_RESPONSE_JSON = example_responses.LIST_MACHINES_RESPONSE
    EXPECTED_STDOUT = """+-----------+---------------+------------------------------------------------------------------------------------+-----+------+-----------+--------------+------------------+
| ID        | Name          | OS                                                                                 | CPU | GPU  | RAM       | State        | Region           |
+-----------+---------------+------------------------------------------------------------------------------------+-----+------+-----------+--------------+------------------+
| psclbvqpc | keton2        | None                                                                               | 1   | None | None      | provisioning | None             |
| psbtuwfvt | New Machine 1 | Ubuntu 18.04.1 LTS; uname: 4.15.0-38-generic; distro: ubuntu; major: 18; minor: 04 | 1   | None | 536870912 | off          | East Coast (NY2) |
+-----------+---------------+------------------------------------------------------------------------------------+-----+------+-----------+--------------+------------------+
"""

    COMMAND_WITH_ALL_OPTIONS = [
        "machines", "list",
        "--agentType", "some_agent_type",
        "--apiKey", "some_key",
        "--autoSnapshotFrequency", "hour",
        "--autoSnapshotSaveCount", "2",
        "--cpus", "8",
        "--dtCreated", "2017-09-23T05:55:00.000Z",
        "--dtLastRun", "2017-09-23T05:55:00.000Z",
        "--gpu", "some_gpu",
        "--id", "some_id",
        "--name", "some_name",
        "--networkId", "some_network_id",
        "--os", "some_os",
        "--performAutoSnapshot", "true",
        "--privateIpAddress", "1.2.3.4",
        "--publicIpAddress", "4.3.2.1",
        "--ram", "123",
        "--region", "CA1",
        "--shutdownTimeoutInHours", "3",
        "--state", "some_state",
        "--storageTotal", "123TB",
        "--storageUsed", "123GB",
        "--teamId", "some_team_id",
        "--updatesPending", "True",
        "--usageRate", "some_usage_rate",
        "--userId", "some_user_id",
    ]
    ALL_OPTIONS_REQUEST_JSON = {
        "params": {
            "agentType": "some_agent_type",
            "autoSnapshotFrequency": "hour",
            "autoSnapshotSaveCount": 2,
            "cpus": 8,
            "dtCreated": "2017-09-23T05:55:00.000Z",
            "dtLastRun": "2017-09-23T05:55:00.000Z",
            "gpu": "some_gpu",
            "machineId": "some_id",
            "name": "some_name",
            "networkId": "some_network_id",
            "os": "some_os",
            "performAutoSnapshot": True,
            "privateIpAddress": "1.2.3.4",
            "publicIpAddress": "4.3.2.1",
            "ram": 123,
            "region": "West Coast (CA1)",
            "shutdownTimeoutInHours": 3,
            "state": "some_state",
            "storageTotal": "123TB",
            "storageUsed": "123GB",
            "teamId": "some_team_id",
            "updatesPending": True,
            "usageRate": "some_usage_rate",
            "userId": "some_user_id",
        }
    }

    COMMAND_WITH_PARAMS_OPTION = ["machines", "list",
                                  "--params", '{"cpus":2,"gpu":"ATIRage128"}']
    REQUEST_JSON_WITH_PARAMS_OPTION = {
        "params": {
            "cpus": 2,
            "gpu": "ATIRage128",
        },
    }

    COMMAND_WITH_OPTIONS_FILE = ["machines", "list", "--optionsFile", ]  # path added in test

    BASIC_COMMAND_WITH_API_KEY = ["machines", "list", "--apiKey", "some_key"]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    EXPECTED_STDOUT_WHEN_NO_MACHINES_WERE_FOUND = "No data found\n"

    COMMAND_WITH_MUTUALLY_EXCLUSIVE_OPTIONS = [
        "machines", "list",
        "--params", '{"cpus":2,"gpu":"ATIRage128"}',
        "--name", "some_name",
    ]

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_machines_list_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_all_options_were_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=self.ALL_OPTIONS_REQUEST_JSON,
                                       params=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_all_options_were_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=self.ALL_OPTIONS_REQUEST_JSON,
                                       params=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_params_option_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_PARAMS_OPTION)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON_WITH_PARAMS_OPTION,
                                       params=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_list_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_list_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_list_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=self.ALL_OPTIONS_REQUEST_JSON,
                                       params=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_list_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_no_machine_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=[], status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_NO_MACHINES_WERE_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=None)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_params_option_was_used_with_mutually_exclusive_option(self, get_patched):
        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_MUTUALLY_EXCLUSIVE_OPTIONS)

        get_patched.assert_not_called()
        assert "You can use either --params dictionary or single filter arguments" in result.output
        assert result.exit_code == 2


class TestRestartMachine(object):
    URL = "https://api.paperspace.io/machines/some_id/restart/"
    COMMAND = [
        "machines", "restart",
        "--id", "some_id",
    ]
    EXPECTED_STDOUT = "Machine restarted\n"

    COMMAND_WITH_API_KEY = [
        "machines", "restart",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["machines", "restart", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Unable to restart instance: Invalid API token\n"

    RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Not found. Please contact support@paperspace.com for help."
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Unable to restart instance: Not found. " \
                                                 "Please contact support@paperspace.com for help.\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_valid_message_when_restart_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_config_file(self, post_patched, machines_restart_config_path):
        post_patched.return_value = MockResponse()
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_restart_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_start_command_was_used_with_invalid_api_token(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_no_content_was_received_in_response(self, post_patched):
        post_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Unable to restart instance\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_machine_with_given_id_was_not_found(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND,
                                                 status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0


class TestShowMachine(object):
    URL = "https://api.paperspace.io/machines/getMachinePublic/"
    BASIC_COMMAND = ["machines", "details", "--id", "some_id"]
    REQUEST_PARAMS = {"machineId": "some_id"}
    EXPECTED_RESPONSE_JSON = example_responses.SHOW_MACHINE_RESPONSE
    EXPECTED_STDOUT = """+---------------------------+------------------------------------------------------------------------------------+
| ID                        | some_id                                                                            |
+---------------------------+------------------------------------------------------------------------------------+
| Name                      | New Machine 1                                                                      |
| OS                        | Ubuntu 18.04.1 LTS; uname: 4.15.0-38-generic; distro: ubuntu; major: 18; minor: 04 |
| RAM                       | 536870912                                                                          |
| CPU                       | 1                                                                                  |
| GPU                       | None                                                                               |
| Storage Total             | 53687091200                                                                        |
| Storage Used              | 110080                                                                             |
| Usage Rate                | C1 Hourly                                                                          |
| Shutdown Timeout In Hours | 1                                                                                  |
| Shutdown Timeout Forces   | False                                                                              |
| Perform Auto Snapshot     | True                                                                               |
| Auto snapshot frequency   | month                                                                              |
| Auto Snapshot Save Count  | 1                                                                                  |
| Agent Type                | LinuxHeadless                                                                      |
| Created                   | 2019-04-11T18:10:29.665Z                                                           |
| State                     | off                                                                                |
| Updates Pending           | False                                                                              |
| Network ID                | nng82wb                                                                            |
| Private IP Address        | 10.64.14.135                                                                       |
| Public IP Address         | None                                                                               |
| Region                    | East Coast (NY2)                                                                   |
| Script ID                 | None                                                                               |
| Last Run                  | None                                                                               |
| Dynamic Public IP         | False                                                                              |
| Last event                | name:     restart                                                                  |
|                           | state:    done                                                                     |
|                           | created:  2019-04-12T12:19:03.814Z                                                 |
+---------------------------+------------------------------------------------------------------------------------+
"""

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "details",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]
    COMMAND_WITH_OPTIONS_FILE = ["machines", "details", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Machine not found"
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Failed to fetch data: Machine not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_machines_list_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output.strip() == self.EXPECTED_STDOUT.strip(), result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_show_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output.strip() == self.EXPECTED_STDOUT.strip()
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_show_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_show_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output.strip() == self.EXPECTED_STDOUT.strip(), result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_list_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_machine_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND,
                                                status_code=404)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0


class TestStartMachine(object):
    URL = "https://api.paperspace.io/machines/some_id/start/"
    COMMAND = [
        "machines", "start",
        "--id", "some_id",
    ]
    EXPECTED_STDOUT = "Machine started\n"

    COMMAND_WITH_OPTIONS_FILE = ["machines", "start", "--optionsFile", ]  # path added in test

    COMMAND_WITH_API_KEY = [
        "machines", "start",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Unable to start instance: Invalid API token\n"

    RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Not found. Please contact support@paperspace.com for help."
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Unable to start instance: Not found. " \
                                                 "Please contact support@paperspace.com for help.\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_valid_message_when_start_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, machines_start_config_path):
        post_patched.return_value = MockResponse()
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_start_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_start_command_was_used_with_invalid_api_token(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_no_content_was_received_in_response(self, post_patched):
        post_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Unable to start instance\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_machine_with_given_id_was_not_found(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND,
                                                 status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0


class TestStopMachine(object):
    URL = "https://api.paperspace.io/machines/some_id/stop/"
    COMMAND = [
        "machines", "stop",
        "--id", "some_id",
    ]
    EXPECTED_STDOUT = "Machine stopped\n"

    COMMAND_WITH_OPTIONS_FILE = ["machines", "stop", "--optionsFile", ]  # path added in test

    COMMAND_WITH_API_KEY = [
        "machines", "stop",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Unable to stop instance: Invalid API token\n"

    RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Not found. Please contact support@paperspace.com for help."
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Unable to stop instance: Not found. " \
                                                 "Please contact support@paperspace.com for help.\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_get_request_and_print_valid_message_when_stop_command_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, machines_stop_config_path):
        post_patched.return_value = MockResponse()
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_stop_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_stop_command_was_used_with_invalid_api_token(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_no_content_was_received_in_response(self, post_patched):
        post_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == "Unable to stop instance\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_machine_with_given_id_was_not_found(self, post_patched):
        post_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_404_MACHINE_NOT_FOUND,
                                                 status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_with(self.URL,
                                        headers=EXPECTED_HEADERS,
                                        json=None,
                                        params=None,
                                        files=None,
                                        data=None)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0


class TestUpdateMachine(object):
    URL = "https://api.paperspace.io/machines/some_id/updateMachinePublic/"
    BASIC_COMMAND = [
        "machines", "update",
        "--id", "some_id",
        "--machineName", "some_name",
    ]
    REQUEST_JSON = {"machineName": "some_name"}

    ALL_COMMANDS = [
        "machines", "update",
        "--id", "some_id",
        "--machineName", "some_name",
        "--shutdownTimeoutInHours", "2",
        "--shutdownTimeoutForces", "true",
        "--performAutoSnapshot", "true",
        "--autoSnapshotFrequency", "hour",
        "--autoSnapshotSaveCount", "1",
        "--dynamicPublicIp", "true",
    ]
    ALL_COMMANDS_REQUEST_JSON = {
        "machineName": "some_name",
        "shutdownTimeoutInHours": 2,
        "shutdownTimeoutForces": True,
        "performAutoSnapshot": True,
        "autoSnapshotFrequency": "hour",
        "autoSnapshotSaveCount": 1,
        "dynamicPublicIp": True,
    }
    COMMAND_WITH_OPTIONS_FILE = ["machines", "update", "--optionsFile", ]  # path added in test

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "update",
        "--id", "some_id",
        "--machineName", "some_name",
        "--apiKey", "some_key",
    ]
    EXPECTED_STDOUT = "Machine updated\n"

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to update resource: Invalid API token\n"

    RESPONSE_JSON_WITH_WRONG_MACHINE_ID = {"error": {"name": "Error", "status": 404, "message": "Not found"}}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machine_create_was_used_with_requested_options(self, get_patched):
        get_patched.return_value = MockResponse({}, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_machine_create_was_used_with_all_options(self, get_patched):
        get_patched.return_value = MockResponse({}, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.ALL_COMMANDS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.ALL_COMMANDS_REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_changed_headers_when_api_key_option_was_used(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_update_config_path):
        get_patched.return_value = MockResponse(example_responses.CREATE_MACHINE_RESPONSE, 200)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_update_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=self.ALL_COMMANDS_REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, 400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_wrong_machine_id_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_MACHINE_ID, 400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.output == "Failed to update resource: Not found\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_error_message_when_no_content_was_received_in_response(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=self.REQUEST_JSON,
                                       params=None,
                                       files=None,
                                       data=None)
        assert result.output == "Failed to update resource\n"
        assert result.exit_code == 0


class TestShowMachineUtilization(object):
    URL = "https://api.paperspace.io/machines/getUtilization/"
    BASIC_COMMAND = [
        "machines", "utilization",
        "--id", "some_id",
        "--billingMonth", "2017-09",
    ]
    REQUEST_PARAMS = {"machineId": "some_id", "billingMonth": "2017-09"}
    EXPECTED_RESPONSE_JSON = example_responses.SHOW_MACHINE_UTILIZATION_RESPONSE
    EXPECTED_STDOUT = """+----------------------+---------------+
| ID                   | some_key      |
+----------------------+---------------+
| Machine Seconds used | 0             |
| Machine Hourly rate  | 0             |
| Storage Seconds Used | 256798.902394 |
| Storage Monthly Rate | 5.00          |
+----------------------+---------------+
"""

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "utilization",
        "--id", "some_id",
        "--billingMonth", "2017-09",
        "--apiKey", "some_key",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["machines", "utilization", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Machine not found"
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Failed to fetch data: Machine not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_machines_utilizaation_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_utilization_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_utilization_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_utilization_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_utilization_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_machine_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND,
                                                status_code=404)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0


class TestWaitForMachine(object):
    URL = "https://api.paperspace.io/machines/getMachinePublic/"
    BASIC_COMMAND = [
        "machines", "waitfor",
        "--id", "some_id",
        "--state", "off",
    ]
    REQUEST_PARAMS = {"machineId": "some_id"}
    EXPECTED_RESPONSE_JSON = example_responses.SHOW_MACHINE_RESPONSE
    EXPECTED_STDOUT = "Machine state: off\n"

    BASIC_COMMAND_WITH_API_KEY = [
        "machines", "waitfor",
        "--id", "some_id",
        "--state", "off",
        "--apiKey", "some_key",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["machines", "waitfor", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND = {
        "error": {
            "name": "Error",
            "status": 404,
            "message": "Machine not found"
        }
    }
    EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND = "Failed to fetch data: Machine not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_machines_waitfor_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_waitfor_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_machines_waitfor_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, machines_waitfor_config_path):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [machines_waitfor_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_machine_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_MACHINE_WAS_NOT_FOUND,
                                                status_code=404)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == self.EXPECTED_STDOUT_WHEN_MACHINE_WAS_NOT_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.REQUEST_PARAMS)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0



import marshmallow as ma

from .. import models
from ..serializers.base import BaseSchema


class ClusterSchema(BaseSchema):
    MODEL = models.Cluster
    id = ma.fields.String()
    name = ma.fields.String()
    type = ma.fields.String()



import copy

import marshmallow

from . import experiment_dataset
from .base import BaseSchema
from .tag import TagSchema
from .. import models, utils


class BaseExperimentSchema(BaseSchema):
    experiment_type_id = marshmallow.fields.Int(required=True, dump_to="experimentTypeId", load_from="experimentTypeId")
    name = marshmallow.fields.Str(required=True)
    ports = marshmallow.fields.Str()
    workspace_url = marshmallow.fields.Str(dump_to="workspaceUrl", load_from="workspaceUrl")
    workspace_ref = marshmallow.fields.Str(dump_to="workspaceRef", load_from="workspaceRef")
    workspace_username = marshmallow.fields.Str(dump_to="workspaceUsername", load_from="workspaceUsername")
    workspace_password = marshmallow.fields.Str(dump_to="workspacePassword", load_from="workspacePassword")
    datasets = marshmallow.fields.Nested(experiment_dataset.ExperimentDatasetSchema, many=True)
    working_directory = marshmallow.fields.Str(dump_to="workingDirectory", load_from="workingDirectory")
    artifact_directory = marshmallow.fields.Str(dump_to="artifactDirectory", load_from="artifactDirectory")
    cluster_id = marshmallow.fields.String(dump_to="clusterId", load_from="clusterId")
    experiment_env = marshmallow.fields.Dict(dump_to="experimentEnv", load_from="experimentEnv")
    project_id = marshmallow.fields.Str(required=True, dump_to="projectHandle", load_from="project_handle")
    model_type = marshmallow.fields.Str(dump_to="modelType", load_from="modelType")
    model_path = marshmallow.fields.Str(dump_to="modelPath", load_from="modelPath")
    is_preemptible = marshmallow.fields.Bool(dump_to="isPreemptible", load_from="isPreemptible")
    id = marshmallow.fields.Str(load_from="handle")
    state = marshmallow.fields.Int()
    tags = marshmallow.fields.Nested(TagSchema, only="name", many=True, load_only=True)

    dt_created = marshmallow.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = marshmallow.fields.DateTime(dump_to="dtModified", load_from="dtModified")
    dt_started = marshmallow.fields.DateTime(dump_to="dtStarted", load_from="dtStarted")
    dt_stopped = marshmallow.fields.DateTime(dump_to="dtStopped", load_from="dtStopped")
    dt_deleted = marshmallow.fields.DateTime(dump_to="dtDeleted", load_from="dtDeleted")

    def get_instance(self, obj_dict, many=False):
        # without popping these marshmallow wouldn't use load_from
        obj_dict.pop("id", None)
        obj_dict.pop("project_id", None)

        ports = obj_dict.get("ports")
        if isinstance(ports, int):
            obj_dict["ports"] = str(ports)

        instance = super(BaseExperimentSchema, self).get_instance(obj_dict, many=many)
        return instance


class SingleNodeExperimentSchema(BaseExperimentSchema):
    MODEL = models.SingleNodeExperiment

    container = marshmallow.fields.Str(required=True, load_from="worker_container")
    machine_type = marshmallow.fields.Str(required=True, dump_to="machineType", load_from="worker_machine_type")
    command = marshmallow.fields.Str(required=True, load_from="worker_command")
    container_user = marshmallow.fields.Str(dump_to="containerUser", load_from="containerUser")
    registry_username = marshmallow.fields.Str(dump_to="registryUsername", load_from="registryUsername")
    registry_password = marshmallow.fields.Str(dump_to="registryPassword", load_from="registryPassword")
    registry_url = marshmallow.fields.Str(dump_to="registryUrl", load_from="registryUrl")

    @marshmallow.pre_dump
    def preprocess(self, data, **kwargs):
        data = copy.copy(data)

        utils.base64_encode_attribute(data, "command")
        return data


class MultiNodeExperimentSchema(BaseExperimentSchema):
    MODEL = models.MultiNodeExperiment

    worker_container = marshmallow.fields.Str(required=True, dump_to="workerContainer", load_from="workerContainer")
    worker_machine_type = marshmallow.fields.Str(required=True, dump_to="workerMachineType",
                                                 load_from="workerMachineType")
    worker_command = marshmallow.fields.Str(required=True, dump_to="workerCommand", load_from="workerCommand")
    worker_count = marshmallow.fields.Int(required=True, dump_to="workerCount", load_from="workerCount")
    parameter_server_container = marshmallow.fields.Str(required=True, dump_to="parameterServerContainer",
                                                        load_from="parameterServerContainer")
    parameter_server_machine_type = marshmallow.fields.Str(required=True, dump_to="parameterServerMachineType",
                                                           load_from="parameterServerMachineType")
    parameter_server_command = marshmallow.fields.Str(required=True, dump_to="parameterServerCommand",
                                                      load_from="parameterServerCommand")
    parameter_server_count = marshmallow.fields.Int(required=True, dump_to="parameterServerCount",
                                                    load_from="parameterServerCount")
    worker_container_user = marshmallow.fields.Str(dump_to="workerContainerUser", load_from="workerContainerUser")
    worker_registry_username = marshmallow.fields.Str(dump_to="workerRegistryUsername",
                                                      load_from="workerRegistryUsername")
    worker_registry_password = marshmallow.fields.Str(dump_to="workerRegistryPassword",
                                                      load_from="workerRegistryPassword")
    worker_registry_url = marshmallow.fields.Str(dump_to="workerRegistryUrl",
                                                 load_from="workerRegistryUrl")
    parameter_server_container_user = marshmallow.fields.Str(required=True, dump_to="parameterServerContainerUser",
                                                             load_from="parameterServerContainerUser")
    parameter_server_registry_username = marshmallow.fields.Str(dump_to="parameterServerRegistryUsername",
                                                                load_from="parameterServerRegistryUsername")
    parameter_server_registry_password = marshmallow.fields.Str(dump_to="parameterServerRegistryPassword",
                                                                load_from="parameterServerRegistryPassword")
    parameter_server_registry_url = marshmallow.fields.Str(dump_to="parameterServerRegistryUrl",
                                                           load_from="parameterServerRegistryUrl")

    @marshmallow.pre_dump
    def preprocess(self, data, **kwargs):
        data = copy.copy(data)

        utils.base64_encode_attribute(data, "worker_command")
        utils.base64_encode_attribute(data, "parameter_server_command")
        return data


class MpiMultiNodeExperimentSchema(BaseExperimentSchema):
    MODEL = models.MpiMultiNodeExperiment

    worker_container = marshmallow.fields.Str(required=True, dump_to="workerContainer", load_from="workerContainer")
    worker_machine_type = marshmallow.fields.Str(required=True, dump_to="workerMachineType",
                                                 load_from="workerMachineType")
    worker_command = marshmallow.fields.Str(required=True, dump_to="workerCommand", load_from="workerCommand")
    worker_count = marshmallow.fields.Int(required=True, dump_to="workerCount", load_from="workerCount")
    master_container = marshmallow.fields.Str(required=True, dump_to="masterContainer",
                                              load_from="masterContainer")
    master_machine_type = marshmallow.fields.Str(required=True, dump_to="masterMachineType",
                                                 load_from="masterMachineType")
    master_command = marshmallow.fields.Str(required=True, dump_to="masterCommand", load_from="masterCommand")
    master_count = marshmallow.fields.Int(required=True, dump_to="masterCount", load_from="masterCount")
    worker_container_user = marshmallow.fields.Str(dump_to="workerContainerUser", load_from="workerContainerUser")
    worker_registry_username = marshmallow.fields.Str(dump_to="workerRegistryUsername",
                                                      load_from="workerRegistryUsername")
    worker_registry_password = marshmallow.fields.Str(dump_to="workerRegistryPassword",
                                                      load_from="workerRegistryPassword")
    worker_registry_url = marshmallow.fields.Str(dump_to="workerRegistryUrl",
                                                 load_from="workerRegistryUrl")
    master_container_user = marshmallow.fields.Str(required=True, dump_to="masterContainerUser",
                                                   load_from="masterContainerUser")
    master_registry_username = marshmallow.fields.Str(dump_to="masterRegistryUsername",
                                                      load_from="masterRegistryUsername")
    master_registry_password = marshmallow.fields.Str(dump_to="masterRegistryPassword",
                                                      load_from="masterRegistryPassword")
    master_registry_url = marshmallow.fields.Str(dump_to="masterRegistryUrl",
                                                 load_from="masterRegistryUrl")

    @marshmallow.pre_dump
    def preprocess(self, data, **kwargs):
        data = copy.copy(data)

        utils.base64_encode_attribute(data, "worker_command")
        utils.base64_encode_attribute(data, "master_command")
        return data



import os
import sys
from platform import system

import six
from distutils.version import StrictVersion

from gradient.clilogger import CliLogger
from gradient.version import version

if six.PY2:
    import xmlrpclib
else:
    import xmlrpc.client as xmlrpclib

logger = CliLogger()


class PackageNotFoundError(Exception):
    pass


class VersionChecker(object):
    def is_up_to_date(self, module_name, current_version):
        version_in_repository = self.get_version_from_repository(module_name)

        up_to_date = StrictVersion(current_version) >= StrictVersion(version_in_repository)
        return up_to_date, version_in_repository

    def get_version_from_repository(self, module_name, repository_url="http://pypi.python.org/pypi"):
        pypi = xmlrpclib.ServerProxy(repository_url)
        versions = pypi.package_releases(module_name)
        if not versions:
            raise PackageNotFoundError("Package {} not found".format(module_name))

        return versions[0]


class GradientVersionChecker(object):
    @classmethod
    def look_for_new_version_with_timeout(cls):
        if not cls._should_check_version():
            return

        if not system() == "Linux":
            cls.look_for_new_version()
            return

        import signal

        class TimeoutError(Exception):
            pass

        def handler(signum, frame):
            raise TimeoutError

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(1)

        try:
            cls.look_for_new_version()
        except TimeoutError:
            pass

        signal.alarm(0)

    @staticmethod
    def look_for_new_version():
        vc = VersionChecker()
        try:
            up_to_date, version_from_repository = vc.is_up_to_date("gradient", version)
        except Exception as e:
            logger.debug(e)
            return

        if not up_to_date:
            msg = "Warning: this version of the Gradient CLI ({current_version}) is out of date. " \
                  "Some functionality might not be supported until you upgrade. \n\n" \
                  "Run `pip install -U gradient` to upgrade\n".format(current_version=version)
            logger.warning(msg)

    @staticmethod
    def _should_check_version():
        if not hasattr(sys.stdin, "isatty"):
            return False
        if not sys.stdin.isatty() or not sys.stdout.isatty():
            return False
        if os.getenv("PAPERSPACE_CLI_DEV") == "true":
            return False

        return True



import pytest
from click import BadParameter

from gradient.commands.experiments import BaseCreateExperimentCommandMixin


class TestHandleDatasetDataMethod(object):
    def test_should_transform_lists_of_datasets_parameters_to_list_of_dicts_and_leave_other_data_intact(self):
        input_data = {
            "some_key": "some_value",
            "dataset_uri_list": ["uri1", "uri2"],
            "dataset_id_list": [None, None, "test"],
            "dataset_name_list": ["name1", None, "name2"],
            "dataset_access_key_id_list": [None, "key_id"],
            "dataset_secret_access_key_list": ["none", "secret"],
            # "dataset_version_id_list": []  # this key does not exist if parameter was not used
            "dataset_etag_list": ["etag"],
            "dataset_volume_kind_list": [],
        }
        expected_data = {
            "some_key": "some_value",
            "datasets": [
                {
                    "uri": "uri1",
                    "id": None,
                    "name": "name1",
                    "aws_access_key_id": None,
                    "aws_secret_access_key": None,
                    "aws_endpoint": None,
                    "version_id": None,
                    "etag": "etag",
                    "volume_kind": None,
                    "volume_size": None,
                },
                {
                    "uri": "uri2",
                    "id": None,
                    "name": None,
                    "aws_access_key_id": "key_id",
                    "aws_secret_access_key": "secret",
                    "aws_endpoint": None,
                    "version_id": None,
                    "etag": None,
                    "volume_kind": None,
                    "volume_size": None,
                },
                {
                    "uri": None,
                    "id": "test",
                    "name": "name2",
                    "aws_access_key_id": None,
                    "aws_secret_access_key": None,
                    "aws_endpoint": None,
                    "version_id": None,
                    "etag": None,
                    "volume_kind": None,
                    "volume_size": None,
                },
            ]

        }

        BaseCreateExperimentCommandMixin._handle_dataset_data(input_data)

        assert expected_data == input_data

    def test_should_raise_exception_if_number_of_any_dataset_parameter_is_greater_than_number_of_dataset_uris(self):
        input_data = {
            "some_key": "some_value",
            "dataset_uri_list": ["uri1", "uri2"],
            "dataset_name_list": ["name1", None],
            "dataset_access_key_id_list": [None, "key_id"],
            "dataset_secret_access_key_list": ["none", "secret"],
            # "dataset_version_id_list": []  # this key does not exist if parameter was not used
            "dataset_etag_list": ["etag", "etag2", "etag3"],  # more etags than uris
        }

        with pytest.raises(BadParameter):
            BaseCreateExperimentCommandMixin._handle_dataset_data(input_data)



import collections
import mimetypes
import os
import tempfile

from . import sdk_exceptions
from .archivers import ZipArchiver
from .clients import http_client
from .config import config
from .logger import MuteLogger
from .utils import MultipartEncoder, MultipartEncoderWithProgressbar


class S3FileUploader(object):
    DEFAULT_MULTIPART_ENCODER_CLS = MultipartEncoder

    def __init__(self, multipart_encoder_cls=None, logger=None, ps_client_name=None):
        """
        :param type(MultipartEncoder) multipart_encoder_cls:
        :param Logger logger:
        """
        self.multipart_encoder_cls = multipart_encoder_cls or self.DEFAULT_MULTIPART_ENCODER_CLS
        self.logger = logger or MuteLogger()
        self.ps_client_name = ps_client_name

    def upload(self, file_path, url, s3_fields=None):
        """Upload a file to S3

        :param str file_path:
        :param str url:
        :param str bucket_name:
        :param dict[str,str] s3_fields:

        """
        # the S3 service requires the file field be the last one in sent object so dict needs to be ordered
        s3_fields = s3_fields or {}
        ordered_s3_fields = collections.OrderedDict(s3_fields)
        with open(file_path, "rb") as file_handle:
            ordered_s3_fields["file"] = (file_path, file_handle)
            multipart_encoder_monitor = self._get_multipart_encoder_monitor(ordered_s3_fields)
            self.logger.debug("Uploading file: {} to url: {}...".format(file_path, url))
            self._upload(url, data=multipart_encoder_monitor)
            self.logger.debug("Uploading completed")

    def _upload(self, url, data):
        """Send data to S3 and raise exception if it was not a success

        :param str url:
        :param encoder.MultipartEncoderMonitor data:
        """
        client = self._get_client(url)
        client.headers = {"Content-Type": data.content_type or ""}
        response = client.post("", data=data)
        if not response.ok:
            raise sdk_exceptions.S3UploadFailedError(response)

    def _get_client(self, url):
        client = http_client.API(url, logger=self.logger, ps_client_name=self.ps_client_name)
        return client

    def _get_multipart_encoder_monitor(self, fields):
        """
        :param dict fields:
        :rtype: encoder.MultipartEncoderMonitor
        """
        multipart_encoder = self.multipart_encoder_cls(fields)
        monitor = multipart_encoder.get_monitor()
        return monitor

    @staticmethod
    def get_bucket_url(bucket_name, s3_fields):
        """
        :param str bucket_name:
        :param dict s3_fields:
        :rtype: str
        """
        s3_object_path = s3_fields["key"]
        url = "s3://{}/{}".format(bucket_name, s3_object_path)
        return url


class S3PutFileUploader(S3FileUploader):
    def _upload(self, url, data):
        """Send data to S3 and raise exception if it was not a success

        :param str url:
        :param encoder.MultipartEncoderMonitor data:
        """
        file_path = data.encoder.fields['file'][0]
        client = self._get_client(url)
        client.headers = {"Content-Type": mimetypes.guess_type(file_path)[0] or ""}

        response = client.put("", data=data)
        if not response.ok:
            raise sdk_exceptions.S3UploadFailedError(response)


class ExperimentFileUploader(object):
    def __init__(self, api_key, uploader=None, logger=None, ps_client_name=None):
        """
        :param str api_key:
        :param S3FileUploader uploader:
        :param Logger logger:
        """
        self.logger = logger or MuteLogger()
        self.experiments_api = http_client.API(
            config.CONFIG_EXPERIMENTS_HOST,
            api_key=api_key,
            logger=self.logger,
            ps_client_name=ps_client_name,
        )
        self.uploader = uploader or S3FileUploader(logger=self.logger, ps_client_name=ps_client_name)

    def upload(self, file_path, project_id, cluster_id=None):
        """Upload file to S3 bucket for a project

        :param str file_path:
        :param str project_id:
        :param str cluster_id:

        :rtype: str
        :return: S3 bucket's URL
        """
        url, bucket_name, fields = self._get_upload_data(file_path, project_id, cluster_id=cluster_id)
        self.uploader.upload(file_path, url, fields)
        bucket_url = self.uploader.get_bucket_url(bucket_name, fields)
        return bucket_url

    def _get_upload_data(self, file_path, project_handle, cluster_id=None):
        """Ask API for data required to upload a file to S3

        :param str file_path:
        :param str project_handle:
        :param str cluster_id:

        :rtype: tuple[str,str,dict]
        :return: URL to which send the file, name of the bucket and a dictionary required by S3 service
        """
        file_name = os.path.basename(file_path)
        params = {"workspaceName": file_name, "projectHandle": project_handle}
        if cluster_id:
            params['clusterHandle'] = cluster_id

        response = self.experiments_api.get("/workspace/get_presigned_url", params=params)
        if response.status_code == 401:
            raise sdk_exceptions.ProjectAccessDeniedError("Access to project denied")
        if response.status_code == 403:
            raise sdk_exceptions.PresignedUrlAccessDeniedError("Access denied")
        if response.status_code == 404:
            raise sdk_exceptions.PresignedUrlUnreachableError("URL not found")
        if not response.ok:
            raise sdk_exceptions.PresignedUrlConnectionError(response.reason)

        try:
            response_data = response.json()
            if response_data["message"] != "success":
                raise sdk_exceptions.PresignedUrlError("Presigned url error: {}".format(response_data))

            url = response_data["data"]["url"]
            bucket_name = response_data["data"]["bucket_name"]
            fields = response_data["data"]["fields"]
        except (KeyError, ValueError):
            raise sdk_exceptions.PresignedUrlMalformedResponseError("Response malformed")

        return url, bucket_name, fields


class S3ProjectFileUploader(ExperimentFileUploader):
    """
    DEPRECATED: This class will be renamed to ExperimentFileUploader in release v0.8
    """
    pass


class S3ModelFileUploader(object):
    DEFAULT_MULTIPART_ENCODER_CLS = MultipartEncoderWithProgressbar

    def __init__(self, api_key, multipart_encoder_cls=None, logger=None, ps_client_name=None, s3uploader=None):
        """
        :param str api_key:
        :param Logger logger:
        """
        self.logger = logger or MuteLogger()
        self.multipart_encoder_cls = multipart_encoder_cls or self.DEFAULT_MULTIPART_ENCODER_CLS
        self.ps_api_client = self._get_client(
            config.CONFIG_HOST,
            api_key=api_key,
            ps_client_name=ps_client_name,
        )
        self.s3uploader = s3uploader or S3PutFileUploader(
            logger=self.logger,
            ps_client_name=ps_client_name,
            multipart_encoder_cls=self.multipart_encoder_cls
        )

    def upload(self, file_path, model_id, cluster_id=None):
        """Upload file to S3 bucket for a project

        :param str file_path:
        :param str model_id:

        :rtype: str
        :return: S3 bucket's URL
        """
        url = self._get_upload_data(file_path, model_id, cluster_id=cluster_id)
        self.s3uploader.upload(file_path, url)
        return url

    def _get_upload_data(self, file_path, model_id, cluster_id=None):
        """Ask API for data required to upload a file to S3

        :param str file_path:
        :param str model_id:

        :rtype: str
        :return: URL to which send the file, name of the bucket and a dictionary required by S3 service
        """
        file_name = os.path.basename(file_path)
        params = {
            "fileName": file_name,
            "modelHandle": model_id,
            "contentType": mimetypes.guess_type(file_path)[0] or "",
        }
        if cluster_id:
            params["clusterId"] = cluster_id

        response = self.ps_api_client.get("/mlModels/getPresignedModelUrl", params=params)
        if not response.ok:
            raise sdk_exceptions.PresignedUrlConnectionError(response.reason)

        try:
            url = response.json()
        except (KeyError, ValueError):
            raise sdk_exceptions.PresignedUrlMalformedResponseError("Response malformed")

        return url

    def _get_client(self, url, ps_client_name=None, api_key=None):
        client = http_client.API(url, logger=self.logger, ps_client_name=ps_client_name, api_key=api_key)
        return client


class S3ModelUploader(S3ModelFileUploader):
    def upload(self, file_path, model_id, cluster_id=None):
        if os.path.isdir(file_path):
            file_path = self._zip_model_directory(file_path)

        return super(S3ModelUploader, self).upload(file_path, model_id, cluster_id=cluster_id)

    def _zip_model_directory(self, dir_path):
        archiver = self._get_archiver()
        archive_path = self._get_archive_path()
        archiver.archive(dir_path, archive_path)
        return archive_path

    def _get_archiver(self):
        return ZipArchiver()

    def _get_archive_path(self):
        archive_file_name = 'model.zip'
        archive_file_path = os.path.join(tempfile.gettempdir(), archive_file_name)
        return archive_file_path


class ExperimentWorkspaceDirectoryUploader(object):
    def __init__(self, api_key, temp_dir=None, archiver=None, project_uploader=None, ps_client_name=None):
        """
        :param str api_key:
        :param str temp_dir:
        :param ZipArchiver archiver:
        :param ExperimentFileUploader project_uploader:
        """
        self.temp_dir = temp_dir or tempfile.gettempdir()
        self.archiver = archiver or ZipArchiver()
        self.ps_client_name = ps_client_name
        self.project_uploader = project_uploader or ExperimentFileUploader(api_key, ps_client_name=ps_client_name)

    def upload(self, workspace_dir_path, project_id, exclude=None, temp_file_name="temp.zip"):
        """Archive and upload a workspace directory

        :param str workspace_dir_path:
        :param str project_id:
        :param list|tuple|None exclude:
        :param str temp_file_name:

        :rtype: str
        :return: URL to the S3 bucket
        """
        archive_path = self.get_archive_path(temp_file_name)
        self.archiver.archive(workspace_dir_path, archive_path, exclude=exclude)
        bucket_url = self.project_uploader.upload(archive_path, project_id)
        return bucket_url

    def get_archive_path(self, temp_file_name):
        archive_path = os.path.join(self.temp_dir, temp_file_name)
        return archive_path


class S3WorkspaceDirectoryUploader(ExperimentWorkspaceDirectoryUploader):
    """
    DEPRECATED: This class will be renamed to ExperimentWorkspaceDirectoryUploader in release v0.8
    """
    pass


class DeploymentWorkspaceDirectoryUploader(object):
    def __init__(self, api_key, uploader=None, logger=None, ps_client_name=None):
        """
        :param str api_key:
        :param S3PutFileUploader uploader:
        :param Logger logger:
        """
        self.logger = logger or MuteLogger()
        self.ps_api_client = http_client.API(
            config.CONFIG_HOST,
            api_key=api_key,
            logger=self.logger,
            ps_client_name=ps_client_name,
        )
        self.uploader = S3PutFileUploader(logger=self.logger, ps_client_name=ps_client_name)

    def _get_upload_data(self, file_path, project_id, cluster_id=None):
        """Ask API for data required to upload deployment workspace a file to S3

        :param str file_path:
        :param str project_id:
        :param str cluster_id:

        :rtype: str
        :return: URL to which send the file, name of the bucket and a dictionary required by S3 service
        """
        file_name = os.path.basename(file_path)
        params = {
            "fileName": file_name,
            "contentType": mimetypes.guess_type(file_path)[0] or "",
        }
        if project_id:
            params['projectId'] = project_id
        if cluster_id:
            params['clusterHandle'] = cluster_id
        response = self.ps_api_client.get("/deployments/getPresignedDeploymentUrl", params=params)
        if not response.ok:
            raise sdk_exceptions.PresignedUrlConnectionError(response.reason)

        try:
            response_data = response.json()
            presigned_url = response_data['presignedUrl']
            bucket_name = response_data['bucketName']
            workspace_url = response_data['workspaceUrl']
        except (KeyError, ValueError):
            raise sdk_exceptions.PresignedUrlMalformedResponseError("Response malformed")

        return presigned_url, bucket_name, workspace_url

    def upload(self, file_path, project_id=None, cluster_id=None, **kwargs):
        """Upload file to S3 bucket for a project

        :param str file_path:
        :param str project_id:
        :param str cluster_id:

        :rtype: str
        :return: S3 bucket's URL
        """
        url, bucket_name, workspace_url = self._get_upload_data(file_path, project_id, cluster_id)
        self.uploader.upload(file_path, url)
        return workspace_url
